------------------------------------------------------------------------------
-- All code here contains stuff layout stuff, such as button positioning,	--
-- Button locking, scaling and so forth buttondragging is also handled here	--
------------------------------------------------------------------------------
local Button = FlexBar2:NewModule("Button");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings
Button.Version = 1;

-- Create a button mixin
Button.ButtonMixin = {};

-- This is called while loading the mixin
function Button.ButtonMixin:Load()
	FlexBar2:Debug("               Setting up layout");

	-- Create the actual frame, inheriting from SecureActionButtonTemplate. Set a name for it, as sadly, some blizzard functions require a frame name instead of a reference
	local Frame = CreateFrame("CheckButton", "FB2_BTN_" .. self.uid, UIParent, "SecureActionButtonTemplate");
	self.Frame = Frame;
	Frame.Object = self;
	
	-- Strata and sizing, drag registering
	Frame:SetFrameStrata("MEDIUM");
	Frame:SetFrameLevel(1);
	Frame:SetWidth(36);
	Frame:SetHeight(36);

	-- Textures setup
	Frame:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress");
	Frame:SetCheckedTexture("Interface\\Buttons\\CheckButtonHilight");
	Frame:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square");
	Frame:GetHighlightTexture():SetBlendMode("ADD");
	Frame:GetCheckedTexture():SetBlendMode("ADD");

	-- Icon setup
	Frame.Icon = self.Frame:CreateTexture(nil, "BACKGROUND")
	Frame.Icon:SetAllPoints(Frame);

	-- Cooldown ticker setup
	Frame.Cooldown = CreateFrame("Cooldown", nil, self.Frame, "CooldownFrameTemplate");

	-- NormalTexture setup
	self.Frame:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2");
	local NormalTexture = self.Frame:GetNormalTexture();
	NormalTexture:SetWidth(60);
	NormalTexture:SetHeight(60);
	NormalTexture:ClearAllPoints();
	NormalTexture:SetPoint("CENTER", 1, -1);

	-- Button settings
	Frame:RegisterForClicks("AnyUp");
	Frame:EnableMouse(true);
	Frame:SetClampedToScreen(true);
	Frame:SetMovable(true);

	-- Dragger frame
	Frame.Dragger = CreateFrame("Frame");
	local Dragger = Frame.Dragger;
	Dragger.Object = self;
	Dragger:SetFrameStrata("MEDIUM");
	Dragger:SetFrameLevel(3);
	Dragger:SetAllPoints(Frame);
	Dragger.Texture = Dragger:CreateTexture(nil,'Overlay');
	Dragger.Texture:SetTexture(0, 0.6, 0, 0.5);
	Dragger.Texture:SetAllPoints(Dragger);
	Dragger:EnableMouse(true);
	Dragger:SetClampedToScreen(true);
	Dragger:SetMovable(false);

	-- Declare a text field to show the button's name
	local ButtonNameField = self.Frame.Dragger:CreateFontString(nil, "ARTWORK");
	self.ButtonNameField = ButtonNameField;
	ButtonNameField:SetFontObject("NumberFontNormal");
	ButtonNameField:SetJustifyH("CENTER");
	ButtonNameField:SetPoint("CENTER", self.Frame, "CENTER");
	-- Set scripts for dragging
	Dragger:SetScript("OnMouseUp", function(Button, MouseClick) if(MouseClick == "LeftButton") then self:StopDragging(); end end);
	Dragger:SetScript("OnMouseDown", function(Button, MouseClick)
		-- Drag
		if(MouseClick == "LeftButton") then
			-- Start dragging
			self:StartDragging();
		end
	end);
	-- Support to receive actions while button is unlocked
	self.Frame.Dragger:SetScript("OnReceiveDrag", function() self:PlaceAction(); end);
end

-- Register 
function Button.ButtonMixin:Activate()
	self:RegisterScript("OnShow", "CheckLock");
	self:RegisterScript("OnHide", "Lock");
	self.ButtonNameField:SetText(self.Name);
end


-- Deactivator functions, in case the button gets deactivated, hide it
function Button.ButtonMixin:Deactivate()
	if(self.Frame:IsShown()) then self:Hide(); end
	self:UnregisterScript("OnShow", "CheckLock");
	self:UnregisterScript("OnHide", "Lock");
end

-- SetDefaults function, restore everything back to normal
function Button.ButtonMixin:SetDefaults()
	self.InfoTable.Show = true;
	self.InfoTable.Lock = false;
	self.InfoTable.Texture = nil;
	local ButtonNumber = tonumber(self.Name);
	-- Little conditional, if the BUtton's name matches a number, then set its position according to it, this makes everything look prettyer when using numbers as button name
	local StartHeight = UIParent:GetHeight() - UIParent:GetHeight() / 4;
	local StartWidth = UIParent:GetWidth() / 6;
	if(ButtonNumber) then
		self.InfoTable.X = StartWidth + (37 * (ButtonNumber - ((math.ceil(ButtonNumber / 12) - 1) * 12) - 1));
		self.InfoTable.Y = StartHeight - (37 * (math.ceil(ButtonNumber / 12) - 1));
	else
		self.InfoTable.X = StartWidth;
		self.InfoTable.Y = StartHeight;
	end
	self.InfoTable.Scale = 1;
	-- Set default button type to actionbutton, all that stuff is handled by the action module
	self.InfoTable.Type = "Action";
end

-- Apply the saved position, status & scale settings
function Button.ButtonMixin:LoadSettings()
	if(self.InfoTable.Show == true) then
		self:Show();
	else
		self:Hide();
	end
	if(self.InfoTable.Lock == true) then
		self:Lock();
	else
		self:Unlock();
	end
	self:SetScale(self.InfoTable.Scale or 1);
	self:SetAlpha(self.InfoTable.Alpha or 1);
	self:SetPosition(self.InfoTable.X, self.InfoTable.Y);
end

-- Change button type + trigger event for it
function Button.ButtonMixin:SetType(Type)
	self.InfoTable.Type = Type;
	self:DispatchEvent("TypeChanged");
end

-- Start dragging method
function Button.ButtonMixin:StartDragging()
	-- Check if the button has a group
	if(type(self.GroupMemberList) == "table") then
		-- Get this frame's positioning
		local MasterX, MasterY = self:GetPosition();
		FlexBar2:Debug("Moving button", self.Name, "StartPos: ", MasterX, MasterY);
		-- Loop through all group members
		for _, GroupMember in pairs(self.GroupMemberList) do
			-- Only do something if the group member is not this button, this is important because we cant anchor a button to itself
			if(GroupMember ~= self) then
				local SlaveFrame = GroupMember.Frame;
				-- Get the group member's position
				SlaveX, SlaveY = GroupMember:GetPosition();
				local Scale = SlaveFrame:GetScale();
				local NewX = (SlaveX - MasterX) / Scale;
				local NewY = (SlaveY - MasterY) / Scale;
				-- Set the member's position relative to this frame
				SlaveFrame:ClearAllPoints();
				SlaveFrame:SetPoint("BOTTOMLEFT", self.Frame , "BOTTOMLEFT", NewX, NewY);
				FlexBar2:Debug("   Dragging ", GroupMember.Name, " along anchored to", NewX, NewY);
			end
		end
	end
	-- DRAG !
	self.Frame:StartMoving();
end

function Button.ButtonMixin:StopDragging()
	-- Check if the button has a group
	if(type(self.GroupMemberList) == "table") then
		FlexBar2:Debug("Groupstop");
		-- Get this frame's positioning
		local MasterX, MasterY = self:GetPosition();
		FlexBar2:Debug("stopping move of button", self.Name, "EndPos: ", MasterX, MasterY);
		-- Loop through all group members
		for _, GroupMember in pairs(self.GroupMemberList) do
			-- Only do something if the group member is not this button, this is important because we cant anchor a button to itself
			if(GroupMember ~= self) then
				local SlaveFrame = GroupMember.Frame;
				-- Get the group member's positioning
				local NewX, NewY = GroupMember:GetPosition();
				SlaveFrame.Object:SetPosition(NewX, NewY);
				SlaveFrame.Object.InfoTable.X = NewX;
				SlaveFrame.Object.InfoTable.Y = NewY;
				FlexBar2:Debug("Stopping dragging for [", _ , "], reanchorred to: ", NewX, NewY);
			end
		end
	end
	-- STOP DRAGGING !
	self.Frame:StopMovingOrSizing();
	-- Lets save
	local X, Y = self:GetPosition();
	self.InfoTable.X = X;
	self.InfoTable.Y = Y;
	self:SetPosition();
end

function Button.ButtonMixin:FrameToCursor()
	local X, Y = GetCursorPosition();
	local UIScale = UIParent:GetScale();
	-- Apply UIScale and get the coords relative to BOTTOMLEFT
	X = X / UIScale;
	Y = Y / UIScale;
	self:SetPosition(X, Y);
	return X, Y;
end

function Button.ButtonMixin:SetPosition(X, Y)
	if(type(X) ~= "number") then X = self.InfoTable.X end
	if(type(Y) ~= "number") then Y = self.InfoTable.Y end
	local Scale = self:GetScale();
	self.Frame:ClearAllPoints();
	self.Frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", X / Scale, Y / Scale);
end

function Button.ButtonMixin:GetPosition()
	-- Always try to return a position relative to TOPLEFT
	local Frame = self.Frame;
	local Scale = self:GetScale();
	return Frame:GetLeft() * Scale, Frame:GetBottom() * Scale;
end

-- Show & hide only do showstates changing, all the rest is done OnShow & OnHide, this is to support showing & hiding due to attribute changes correctly
function Button.ButtonMixin:Show()
	-- Showstats has to be set to actualy show the button in all states, if we dont do this there is a chance it gets hidden on a state change
	self:SetAttribute("showstates", "*", false);
	self.Frame:Show();
end

function Button.ButtonMixin:Hide()
	-- Showstates has to be set to hide the button in all states, if we dont do this there is a chance it gets shown on a state change
	self:SetAttribute("showstates", "!*", false);
	self.Frame:Hide();
end

function Button.ButtonMixin:CheckLock()
	-- Restore lock settings
	if(self.InfoTable.Lock == true) then
		self:Lock();
	else
		self:Unlock();
	end
end

function Button.ButtonMixin:Lock()
	self.Frame.Dragger:Hide();
end

function Button.ButtonMixin:Unlock()
	if(self.Frame:IsShown()) then
		self.Frame.Dragger:Show();
	end
end

function Button.ButtonMixin:GetIcon(Icon)
	return self.Frame.Icon:GetTexture(Icon);
end

function Button.ButtonMixin:SetIcon(Icon, IgnoreSavedTexture)
	self.Frame.Icon:SetTexture(IgnoreSavedTexture and Icon or (self.InfoTable.Texture or Icon));
end

function Button.ButtonMixin:SetScale(Scale)
	self.Frame:SetScale(Scale);
	-- Reposition the button with the new scale
	FlexBar2:Debug("Repositioning button with new scale");
	self:SetPosition();
end

-- Gets scale
function Button.ButtonMixin:GetScale()
	return self.Frame:GetScale();
end

function Button.ButtonMixin:SetAlpha(Alpha)
	self.Frame:SetAlpha(Alpha);
end

-- Gets scale
function Button.ButtonMixin:GetAlpha()
	return self.Frame:GetAlpha();
end


-- Small group mixin for showing the group name above it's first button
Button.GroupMixin = {};

function Button.GroupMixin:Activate(Name, ButtonTable)
	local ButtonList = FlexBar2.Buttons;
	local GroupNameField = ButtonList[ButtonTable[1]].Frame.Dragger:CreateFontString(nil, "ARTWORK");
	GroupNameField:SetFontObject("NumberFontNormal");
	GroupNameField:SetJustifyH("CENTER");
	GroupNameField:SetPoint("CENTER", ButtonList[ButtonTable[1]].Frame.Dragger, "CENTER", 0, 20);
	GroupNameField:SetText(Name);
	self.GroupNameField = GroupNameField;
end

function Button.GroupMixin:Deactivate()
	self.GroupNameField:Hide();
end
